using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;
using Microsoft.DirectX.Direct3D;

namespace BenLincoln.TheLostWorlds.CDTextures
{
    public enum TextureFileType
    {
        Unknown,
        SoulReaverPlaystation,
        SoulReaverPC,
        SoulReaverDreamcast,
        SoulReaver2Playstation2,
        SoulReaver2PC
    }

    public abstract class TextureFile
    {
        // Fields
        protected bool _ErrorOccurred = false;
        protected FileInfo _FileInfo;
        protected string _FilePath;
        protected TextureFileType _FileType;
        protected string _FileTypeName;
        protected string _LastErrorMessage = "";
        protected int _TextureCount;

        // Methods
        public TextureFile(string path)
        {
            this._FilePath = path;
            if (File.Exists(path))
            {
                try
                {
                    this._FileInfo = new FileInfo(path);
                }
                catch (IOException exception)
                {
                    throw new IOException("Unable to access the file '" + path + "'", exception);
                }
            }
        }

        protected abstract Texture _GetTexture(Device device, int index);
        protected abstract Bitmap _GetTextureAsBitmap(int index);
        protected abstract int _GetTextureCount();
        public virtual void ExportFile(int index, string outPath)
        {
            this._ErrorOccurred = false;
            this._LastErrorMessage = "";
            try
            {
                MemoryStream dataAsStream = this.GetDataAsStream(index);
                BinaryReader reader = new BinaryReader(dataAsStream);
                FileStream output = new FileStream(outPath, FileMode.Create, FileAccess.Write);
                BinaryWriter writer = new BinaryWriter(output);
                for (int i = 0; i < dataAsStream.Length; i++)
                {
                    writer.Write(reader.ReadByte());
                }
                writer.Close();
                output.Close();
                reader.Close();
                dataAsStream.Close();
            }
            catch (Exception exception)
            {
                this._ErrorOccurred = true;
                this._LastErrorMessage = exception.Message;
            }
        }

        public virtual void ExportFileThreaded(object parms)
        {
            try
            {
                ArrayList list = (ArrayList)parms;
                this.ExportFile((int)list[0], (string)list[1]);
            }
            catch (InvalidCastException)
            {
                throw new InvalidCastException("Passed object must be an arraylist with the first element being an int and the second a string.");
            }
        }

        public abstract MemoryStream GetDataAsStream(int index);
        public static TextureFileType GetFileType(string path)
        {
            switch (Path.GetExtension(path).ToLower())
            {
                //case ".vrm":
                //    return VRMTextureFile.GetVRMType(path);

                case ".big":
                    return TextureFileType.SoulReaverPC;

                case ".vq":
                    return TextureFileType.SoulReaverDreamcast;

                case ".crm":
                    return TextureFileType.SoulReaverPlaystation;
            }
            return TextureFileType.Unknown;
        }

        public Texture GetTexture(Device device, int index)
        {
            if (index < 0)
            {
                throw new IndexOutOfRangeException("Texture index cannot be negative.");
            }
            if (index > (this._TextureCount - 1))
            {
                throw new IndexOutOfRangeException("Texture " + index.ToString() + " was requested but the file '" + this._FilePath + "' only contains enough data for " + this._TextureCount.ToString() + " textures.");
            }
            return this._GetTexture(device, index);
        }

        public Bitmap GetTextureAsBitmap(int index)
        {
            if (index < 0)
            {
                throw new IndexOutOfRangeException("Texture index cannot be negative.");
            }
            if (index > (this._TextureCount - 1))
            {
                throw new IndexOutOfRangeException("Texture " + index.ToString() + " was requested but the file '" + this._FilePath + "' only contains enough data for " + this._TextureCount.ToString() + " textures.");
            }
            return this._GetTextureAsBitmap(index);
        }

        // Properties
        public bool ErrorOccurred
        {
            get
            {
                return this._ErrorOccurred;
            }
        }

        public FileInfo FileInfo
        {
            get
            {
                return this._FileInfo;
            }
        }

        public string FilePath
        {
            get
            {
                return this._FilePath;
            }
        }

        public TextureFileType FileType
        {
            get
            {
                return this._FileType;
            }
        }

        public string FileTypeName
        {
            get
            {
                return this._FileTypeName;
            }
        }

        public string LastErrorMessage
        {
            get
            {
                return this._LastErrorMessage;
            }
        }

        public int TextureCount
        {
            get
            {
                return this._TextureCount;
            }
        }
    }
}
