using System;
using System.Collections.Generic;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    public interface ModelExObject : ModelExFile
    {
        void Load(String strFilePath);
        void Save(String strFilePath, ModelFormat eFormat);
    }

    class ModelExObjectFile : ModelExObject
    {
        protected ModelExSystem m_xSystem;
        protected String m_strFileName;

        public WeakReference Parent { get { return null; } }
        public String FileName { get { return m_strFileName; } }

        protected ModelExObjectFile(ModelExSystem xSystem, String strFilePath)
        {
            m_xSystem = xSystem;
            m_strFileName = strFilePath;

            if (m_xSystem == null)
            {
                throw new Exception("Object file must have a system");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Object file must have a folder");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Object file must have a name");
            }

            Load(strFilePath);
        }

        ~ModelExObjectFile()
        {
            m_xSystem = null;
            m_strFileName = null;
        }

        public static ModelExObjectFile Create(ModelExSystem xSystem, String strFilePath)
        {
            ModelExObjectFile xObjectFile = xSystem.FindObjectFile(strFilePath);

            if (xObjectFile == null)
            {
                xObjectFile = new ModelExObjectFile(xSystem, strFilePath);
                xSystem.AddObjectFile(xObjectFile);
            }

            return xObjectFile;
        }

        public virtual void Load(String strFilePath)
        {
        }

        public virtual void Save(String strFilePath, ModelFormat eFormat)
        {
        }
    }

    class ModelExObjectInstance
    {
        class StateBackup
        {
            private Device m_xDevice;

            public readonly Cull m_eCullMode;
            public readonly FillMode m_eFillMode;
            public readonly bool m_bLighting;
            public readonly TextureFilter m_eMagFilter;
            public readonly TextureFilter m_eMinFilter;
            public readonly Matrix m_xWorld;

            public StateBackup(Device xDevice)
            {
                m_xDevice = xDevice;

                m_eCullMode = m_xDevice.RenderState.CullMode;
                m_eFillMode = m_xDevice.RenderState.FillMode;
                m_bLighting = m_xDevice.RenderState.Lighting;
                m_eMagFilter = m_xDevice.SamplerState[0].MagFilter;
                m_eMinFilter = m_xDevice.SamplerState[0].MinFilter;
                m_xWorld = m_xDevice.Transform.World;
            }

            public void Restore()
            {
                m_xDevice.RenderState.CullMode = m_eCullMode;
                m_xDevice.RenderState.FillMode = m_eFillMode;
                m_xDevice.RenderState.Lighting = m_bLighting;
                m_xDevice.SamplerState[0].MagFilter = m_eMagFilter;
                m_xDevice.SamplerState[0].MinFilter = m_eMinFilter;
                m_xDevice.Transform.World = m_xWorld;
            }
        }

        protected ModelExSystem m_xSystem;
        protected String m_strFileName;

        protected ModelExObjectFile m_xFile;
        protected ModelExModelList m_xModels;
        protected ModelExEffectList m_xEffects;
        protected ModelExTextureList m_xTextures;
        protected List<ModelExContainer> m_xContainers;

        protected int m_iModelIndex;
        protected Quaternion m_xRotation;
        protected ModelExArcBall m_xArcBall;
        protected Vector3 m_xPosition;
        protected Vector3 m_xScale;

        public String FileName { get { return m_strFileName; } }
        public ModelExModelList Models { get { return m_xModels; } }
        public ModelExTextureList Textures { get { return m_xTextures; } }
        public ModelExEffectList Effects { get { return m_xEffects; } }
        public virtual String TextureArchives { get { return null; } }
        public virtual String TextureArchiveExt { get { return null; } }
        public int ModelIndex
        {
            get { return m_iModelIndex; }
            set
            {
                if (Models != null)
                {
                    m_iModelIndex = value;
                    m_iModelIndex = Math.Max(0, m_iModelIndex);
                    m_iModelIndex = Math.Min(m_iModelIndex, Models.Count);
                    m_xScale = m_xModels.List[m_iModelIndex].Frame.ModelExContainer.Scale;
                }
                else
                {
                    m_iModelIndex = 0;
                    m_xScale = new Vector3(1.0f, 1.0f, 1.0f);
                }
            }
        }
        public Quaternion Rotation
        {
            get { return m_xRotation; }
            set
            {
                m_xRotation = value;
                if (m_xArcBall != null)
                {
                    m_xArcBall.Rotation = m_xRotation;
                }
            }
        }
        public ModelExArcBall ArcBall
        {
            get { return m_xArcBall; }
            set { m_xArcBall = value; }
        }

        public ModelExObjectInstance(ModelExSystem xSystem, String strFileName)
        {
            m_xSystem = xSystem;

            try
            {
                m_strFileName = Path.GetFileName(strFileName);
            }
            catch { }

            if (xSystem == null)
            {
                throw new Exception("Object instance must have a system");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Object instance must have a file name");
            }

            m_xModels = new ModelExModelList();
            m_xEffects = new ModelExEffectList();
            m_xTextures = new ModelExTextureList();
            m_xContainers = new List<ModelExContainer>();

            m_xRotation = Quaternion.Identity;
            m_xPosition = Vector3.Empty;
            m_xScale = new Vector3(1.0f, 1.0f, 1.0f);
            m_iModelIndex = 0;
        }

        ~ModelExObjectInstance()
        {
            m_xFile = null;
        }

        public void PrepareFilesForSave(bool bUseEffects)
        {
            for (int c = 0; c < m_xContainers.Count; c++)
            {
                // Update effect names
                EffectInstance[] axOldEffects = m_xContainers[c].GetEffectInstancesBackup();
                EffectInstance[] axNewEffects = new EffectInstance[axOldEffects.Length];
                int[] aiEffectIDs = m_xContainers[c].Effects;
                for (int e = 0; e < axOldEffects.Length; e++)
                {
                    if (aiEffectIDs[e] < m_xEffects.Count)
                    {
                        axNewEffects[e] = axOldEffects[e];
                        axNewEffects[e].EffectFilename = m_xEffects[aiEffectIDs[e]].Name;
                    }
                    else
                    {
                        aiEffectIDs[e] = -1;
                        axNewEffects[e].EffectFilename = "";
                    }
                }

                m_xContainers[c].SetEffectInstancesBackup(axNewEffects);

                if (bUseEffects)
                {
                    m_xContainers[c].SetEffectInstances(axNewEffects);
                }
                else
                {
                    m_xContainers[c].SetEffectInstances(null);
                }

                // Update texture names
                ExtendedMaterial[] axOldMaterials = m_xContainers[c].GetMaterials();
                ExtendedMaterial[] axNewMaterials = new ExtendedMaterial[axOldMaterials.Length];
                int[] aiTextureIDs = m_xContainers[c].Textures;
                for (int m = 0; m < axOldMaterials.Length; m++)
                {
                    if (aiTextureIDs[m] < m_xTextures.Count)
                    {
                        axNewMaterials[m] = axOldMaterials[m];
                        axNewMaterials[m].TextureFilename = m_xTextures[aiTextureIDs[m]].Name;
                    }
                    else
                    {
                        aiTextureIDs[m] = -1;
                        axNewMaterials[m].TextureFilename = "";
                    }
                }
                m_xContainers[c].SetMaterials(axNewMaterials);
            }
        }

        public void CleanupFilesAfterSave(bool bUseEffects)
        {
            for (int c = 0; c < m_xContainers.Count; c++)
            {
                EffectInstance[] axEffects = m_xContainers[c].GetEffectInstancesBackup();
                if (bUseEffects)
                {
                    m_xContainers[c].SetEffectInstances(axEffects);
                }
                else
                {
                    m_xContainers[c].SetEffectInstances(null);
                }
            }
        }

        public virtual void Load(string strFilePath)
        {
            m_xFile = ModelExObjectFile.Create(m_xSystem, strFilePath);

            m_xModels.List.Clear();
            m_xEffects.List.Clear();
            m_xTextures.List.Clear();
            m_xContainers.Clear();

            if (m_xFile == null)
            {
                throw new Exception();
            }

            ModelExModelInstance xModel = new ModelExModelInstance(m_xSystem, m_strFileName);
            xModel.Load(strFilePath);

            float fScale = 1.0f / xModel.File.Containers[0].Radius;
            xModel.File.Containers[0].Scale = new Vector3(fScale, fScale, fScale);

            m_xModels.List.Add(xModel);
            m_xContainers.AddRange(xModel.File.Containers);

            for (int c = 0; c < m_xContainers.Count; c++)
            {
                // Update effects
                EffectInstance[] axEffectInstances = m_xContainers[c].GetEffectInstancesBackup();
                for (int iIndex = 0; iIndex < m_xContainers[c].Mesh.NumberAttributes; iIndex++)
                {
                    m_xContainers[c].Effects[iIndex] = -1;
                    for (int e = 0; e < m_xEffects.Count; e++)
                    {
                        if (m_xEffects[e].Name == axEffectInstances[iIndex].EffectFilename)
                        {
                            m_xContainers[c].Effects[iIndex] = e;
                        }
                    }

                    if (m_xContainers[c].Effects[iIndex] == -1)
                    {
                        ModelExEffectInstance xEffect = new ModelExEffectInstance(m_xSystem, axEffectInstances[iIndex]);
                        m_xEffects.List.Add(xEffect);
                        m_xContainers[c].Effects[iIndex] = m_xEffects.Count - 1;
                    }
                }

                // Update textures
                ExtendedMaterial[] axExtendedMaterials = m_xContainers[c].GetMaterials();
                for (int iIndex = 0; iIndex < m_xContainers[c].Mesh.NumberAttributes; iIndex++)
                {
                    m_xContainers[c].Textures[iIndex] = -1;
                    for (int t = 0; t < m_xTextures.Count; t++)
                    {
                        if (m_xTextures[t].Name == axExtendedMaterials[iIndex].TextureFilename)
                        {
                            m_xContainers[c].Textures[iIndex] = t;
                            break;
                        }
                    }

                    if (m_xContainers[c].Textures[iIndex] == -1)
                    {
                        ModelExTextureInstance xTexture = new ModelExTextureInstance(m_xSystem, axExtendedMaterials[iIndex].TextureFilename);
                        m_xTextures.List.Add(xTexture);
                        m_xContainers[c].Textures[iIndex] = m_xTextures.Count - 1;
                    }
                }
            }

            String strFolderName = Path.GetDirectoryName(strFilePath);
            LoadEffectsFromFolder(strFolderName);
            LoadTexturesFromFolder(strFolderName);

            ModelIndex = 0;
        }

        public virtual void Save(string strFilePath, ModelFormat eFormat)
        {
            if (m_xFile != null)
            {
                m_xFile.Save(strFilePath, eFormat);
            }
        }

        public virtual void LoadEffectsFromFolder(String strFolderName)
        {
            for (int e = 0; e < m_xEffects.Count; e++)
            {
                String strFilePath = null;
                try
                {
                    strFilePath = Path.Combine(strFolderName, m_xEffects[e].Name);
                }
                catch { }
                m_xEffects.List[e].Load(strFilePath);
            }
        }

        public virtual void LoadTexturesFromFile(String strFilePath)
        {
        }

        public virtual void LoadTexturesFromFolder(String strFolderPath)
        {
            for (int t = 0; t < m_xTextures.Count; t++)
            {
                String strFilePath = null;
                try
                {
                    strFilePath = Path.Combine(strFolderPath, m_xTextures[t].Name);
                }
                catch { }
                m_xTextures.List[t].Load(strFilePath);
            }
        }

        public void Render(RenderFlags eRenderFlags, CullType eCullType, float fScale)
        {
            if (Models == null)
            {
                return;
            }
            if (m_xArcBall != null)
            {
                m_xRotation = m_xArcBall.Rotation;
            }

            ModelExFrame xFrame = m_xModels.List[m_iModelIndex];
            ModelExContainer xContainer = xFrame.ModelExContainer;
            EffectInstance[] axEffectInstances = xContainer.GetEffectInstances();
            StateBackup xStateBackup = new StateBackup(m_xSystem.Device);
            for (int e = 0; e < xContainer.Effects.Length; e++)
            {
                Effect xEffect = Effects.List[xContainer.Effects[e]];

                #region Effect Parameters

                EffectDefault[] axEffectDefaults = axEffectInstances[e].GetDefaults();
                foreach (EffectDefault xDefault in axEffectDefaults)
                {
                    EffectHandle xParamHandle = xEffect.GetParameter(null, xDefault.ParameterName);

                    if (xParamHandle == null)
                    {
                        continue;
                    }

                    switch (xDefault.Type)
                    {
                        case EffectDefaultType.Floats:
                            try
                            {
                                xEffect.SetValue(xParamHandle, (float[])xDefault.Data);
                            }
                            catch { }
                            break;
                        case EffectDefaultType.Integer:
                            try
                            {
                                xEffect.SetValue(xParamHandle, (int)xDefault.Data);
                            }
                            catch { }
                            break;
                        case EffectDefaultType.String:
                            try
                            {
                                xEffect.SetValue(xParamHandle, (String)xDefault.Data);
                            }
                            catch
                            {
                                Texture xTexture = null;
                                xEffect.SetValue(xParamHandle, xTexture);
                            };
                            break;
                    }
                }

                #endregion

                EffectHandle xHandle = xEffect.GetTechnique(0);
                if (xHandle == null || !xEffect.IsTechniqueValid(xHandle))
                {
                    throw new Exception("Technique is not valid");
                }

                int iPasses = xEffect.Begin(FX.None);

                for (int p = 0; p < iPasses; p++)
                {
                    xEffect.BeginPass(p);

                    #region Textures
                    if ((eRenderFlags & RenderFlags.Textures) != 0)
                    {
                        Texture xTexture = Textures.List[xContainer.Textures[e]];
                        m_xSystem.Device.SetTexture(0, xTexture);
                    }
                    else
                    {
                        m_xSystem.Device.SetTexture(0, null);
                    }
                    #endregion

                    #region Materials
                    m_xSystem.Device.Material = xContainer.GetMaterials()[e].Material3D;
                    #endregion

                    #region Lighting
                    if ((eRenderFlags & RenderFlags.Lighting) == 0)
                    {
                        m_xSystem.Device.RenderState.Lighting = false;
                    }
                    #endregion

                    #region Texture Filtering
                    if ((eRenderFlags & RenderFlags.TextureFiltering) == 0.0f)
                    {
                        m_xSystem.Device.SamplerState[0].MagFilter = TextureFilter.Point;
                        m_xSystem.Device.SamplerState[0].MinFilter = TextureFilter.Point;
                    }
                    #endregion

                    #region Cull Mode
                    m_xSystem.Device.RenderState.CullMode = (Cull)eCullType;
                    #endregion

                    #region Wireframe Mode
                    if ((eRenderFlags & RenderFlags.WireFrame) != 0.0f)
                    {
                        m_xSystem.Device.RenderState.FillMode = FillMode.WireFrame;
                    }
                    #endregion

                    #region Transforms
                    Vector3 xScale = m_xScale;
                    if ((eRenderFlags & RenderFlags.Scaling) == 0)
                    {
                        xScale = new Vector3(fScale, fScale, fScale);
                    }
                    m_xSystem.Device.Transform.World =
                        Matrix.Translation(m_xPosition - xContainer.Centre) *
                        Matrix.Scaling(xScale) *
                        Matrix.RotationQuaternion(m_xRotation);
                    #endregion

                    #region Rendering
                    xContainer.Mesh.DrawSubset(e);
                    #endregion

                    xEffect.EndPass();
                }

                xEffect.End();
            }

            xStateBackup.Restore();
        }
    }
}
