using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    class ModelExObject
    {
        #region Classes

        protected class ModelExFrame : Frame
        {
        }

        protected class ModelExMeshContainer : MeshContainer
        {
            protected Effect[]          m_axEffects;
            protected bool[]            m_abEffectIsDefault;
            protected Material[]        m_axMaterials;
            protected Texture[]         m_axTextures;
            protected String[]          m_astrTextureNames;

            public Effect[]             Effects
            {
                get { return m_axEffects; }
                set { m_axEffects = value; }
            }
            public bool[]               DefaultEffects
            {
                get { return m_abEffectIsDefault; }
                set { m_abEffectIsDefault = value; }
            }
            public Material[]           Materials
            {
                get { return m_axMaterials; }
                set { m_axMaterials = value; }
            }
            public Texture[]            Textures
            {
                get { return m_axTextures; }
                set { m_axTextures = value; }
            }
            public String[]             TextureNames
            {
                get
                {
                    return m_astrTextureNames;
                }
                set
                {
                    m_astrTextureNames = value;
                }
            }
            public Mesh                 Mesh
            {
                get { return MeshData.Mesh; }
                set
                {
                    MeshData xMeshData = MeshData;
                    xMeshData.Mesh = value;
                    MeshData = xMeshData;
                }
            }
        }

        protected class StateBackup
        {
            private Device m_xDevice;

            public readonly Cull            m_eCullMode;
            public readonly FillMode        m_eFillMode;
            public readonly bool            m_bLighting;
            public readonly TextureFilter   m_eMagFilter;
            public readonly TextureFilter   m_eMinFilter;
            public readonly Matrix          m_xWorld;

            public StateBackup(ref Device xDevice)
            {
                m_xDevice       = xDevice;

                m_eCullMode     = m_xDevice.RenderState.CullMode;
                m_eFillMode     = m_xDevice.RenderState.FillMode;
                m_bLighting     = m_xDevice.RenderState.Lighting;
                m_eMagFilter    = m_xDevice.SamplerState[0].MagFilter;
                m_eMinFilter    = m_xDevice.SamplerState[0].MinFilter;
                m_xWorld        = m_xDevice.Transform.World;
            }

            public void Restore()
            {
                m_xDevice.RenderState.CullMode      = m_eCullMode;
                m_xDevice.RenderState.FillMode      = m_eFillMode;
                m_xDevice.RenderState.Lighting      = m_bLighting;
                m_xDevice.SamplerState[0].MagFilter = m_eMagFilter;
                m_xDevice.SamplerState[0].MinFilter = m_eMinFilter;
                m_xDevice.Transform.World           = m_xWorld;
            }
        }

        #endregion

        protected bool                          m_bInitialised;
        protected String                        m_strFileName;
        protected String                        m_strObjectName;
        protected Device                        m_xDevice;
        // protected ModelExFrame               m_xRootFrame;
        protected ModelExMeshContainer          m_xMeshContainer;
        protected List<ModelExMeshContainer>    m_xMeshContainers;
        protected List<Texture>                 m_xTextures;
        protected int                           m_iModelIndex;

        public int ModelIndex
        {
            get { return m_iModelIndex; }
            set
            {
                if (m_xMeshContainers != null)
                {
                    if (value < 0)
                    {
                        m_iModelIndex = 0;
                    }
                    else if (value < m_xMeshContainers.Count)
                    {
                        m_iModelIndex = value;
                    }
                    m_xMeshContainer = m_xMeshContainers[m_iModelIndex];
                }
                else
                {
                    m_iModelIndex = 0;
                    m_xMeshContainer = null;
                }
            }
        }
        public String[] ModelNames
        {
            get
            {
                if (m_xMeshContainers == null)
                {
                    return null;
                }

                String[] strModelNames = new String[m_xMeshContainers.Count];
                for (int m = 0; m < m_xMeshContainers.Count; m++)
                {
                    strModelNames[m] = m_xMeshContainers[m].Name;
                }

                return strModelNames;
            }
        }
        public String[] TextureNames
        {
            get
            {
                if (m_xMeshContainer != null)
                {
                    return m_xMeshContainer.TextureNames;
                }
                return null;
            }
        }
        public virtual String TextureArchives
        {
            get { return null; }
        }
        public virtual String TextureArchiveExt
        {
            get { return null; }
        }
        public String ObjectName
        {
            get { return m_strObjectName; }
        }

        #region Render states

        protected bool          m_bLighting;
        protected bool          m_bTextures;
        protected bool          m_bTextureFiltering;
        protected Cull          m_eCullMode;
        protected bool          m_bWireframe;

        public bool             Lighting
        {
            get { return m_bLighting; }
            set { m_bLighting = value; }
        }
        public bool             Textures
        {
            get { return m_bTextures; }
            set { m_bTextures = value; }
        }
        public bool             TextureFiltering
        {
            get { return m_bTextureFiltering; }
            set { m_bTextureFiltering = value; }
        }
        public Cull             Cull
        {
            get { return m_eCullMode; }
            set { m_eCullMode = value; }
        }
        public bool             Wireframe
        {
            get { return m_bWireframe; }
            set { m_bWireframe = value; }
        }

        protected virtual TextureOperation  DefaultColorOperation
        {
            get { return TextureOperation.Modulate; }
        }
        protected virtual bool              DefaultLighting
        {
            get { return true; }
        }
        protected virtual ColorSource       DefaultAmbientMaterialSource
        {
            get { return ColorSource.Material; }
        }
        protected virtual ColorSource       DefaultDiffuseMaterialSource
        {
            get { return ColorSource.Color1; }
        }
        protected virtual bool              DefaultColorVertex
        {
            get { return true; }
        }

        #endregion

        #region Orientation

        protected Quaternion                m_xRotation;
        protected Vector3                   m_xPosition;
        protected ModelExArcBall            m_xArcBall;

        public Quaternion                   Rotation
        {
            get { return m_xRotation; }
            set { m_xRotation = value; }
        }
        public Vector3                      Position
        {
            get { return m_xPosition; }
            set { m_xPosition = value; }
        }
        public ModelExArcBall               ArcBall
        {
            get { return m_xArcBall; }
        }

        #endregion

        #region Dimensions

        protected Vector3                   m_xCentre;
        protected float                     m_fRadius;
        protected float                     m_fScale;

        public Vector3                      Centre
        {
            get { return m_xCentre; }
        }
        public float                        Radius
        {
            get { return m_fRadius; }
        }
        public float                        Scale
        {
            get { return m_fScale; }
        }

        #endregion

        public ModelExObject(String strFileName, ref Device xDevice)
        {
            if (!System.IO.File.Exists(strFileName))
            {
                throw new Exception("File does not exist.");
            }

            m_strFileName = strFileName;
            m_xDevice = xDevice;
            m_xMeshContainers = new List<ModelExMeshContainer>();
            m_xTextures = new List<Texture>();

            Load();

            GraphicsStream xVertexData = m_xMeshContainer.Mesh.VertexBuffer.Lock(0, 0, LockFlags.None);
            m_fRadius = Geometry.ComputeBoundingSphere(
                xVertexData,
                m_xMeshContainer.Mesh.NumberVertices,
                m_xMeshContainer.Mesh.VertexFormat,
                out m_xCentre
            );
            m_fScale = 1.0f / m_fRadius;
            m_xMeshContainer.Mesh.VertexBuffer.Unlock();

            m_xPosition = Vector3.Empty;
            m_xRotation = Quaternion.Identity;

            Control xRenderWindow = m_xDevice.PresentationParameters.DeviceWindow;
            m_xArcBall = new ModelExArcBall(ref xRenderWindow, false);
        }

        protected virtual void Load()
        {
            if (!m_bInitialised)
            {
                #region Initialise

                try
                {
                    m_strObjectName = Path.GetFileNameWithoutExtension(m_strFileName);
                }
                catch (Exception)
                {
                    m_strObjectName = "";
                }

                m_xMeshContainer = new ModelExMeshContainer();
                m_xMeshContainer.Name = m_strObjectName;

                ExtendedMaterial[] xExMaterials;
                EffectInstance[] xEffectInstances;

                // Load the mesh
                m_xMeshContainer.Mesh = Mesh.FromFile(
                    m_strFileName, MeshFlags.Managed,
                    m_xDevice, out xExMaterials, out xEffectInstances
                );
                if ((m_xMeshContainer.Mesh.VertexFormat & VertexFormats.Normal) != VertexFormats.Normal)
                {
                    m_xMeshContainer.Mesh = NormalizeMesh(m_xMeshContainer.Mesh);
                }

                // Load the attributes
                m_xMeshContainer.Materials      = new Material[m_xMeshContainer.Mesh.NumberAttributes];
                m_xMeshContainer.Effects        = new Effect[m_xMeshContainer.Mesh.NumberAttributes];
                m_xMeshContainer.DefaultEffects = new bool[m_xMeshContainer.Mesh.NumberAttributes];
                m_xMeshContainer.Textures       = new Texture[m_xMeshContainer.Mesh.NumberAttributes];
                m_xMeshContainer.TextureNames   = new String[m_xMeshContainer.Mesh.NumberAttributes];

                for (int iIndex = 0; iIndex < m_xMeshContainer.Mesh.NumberAttributes; iIndex++)
                {
                    #region Load Material
                    m_xMeshContainer.Materials[iIndex] = xExMaterials[iIndex].Material3D;
                    #endregion
                    
                    #region Load Texture
                    LoadTexture(xExMaterials[iIndex].TextureFilename, iIndex);
                    #endregion

                    #region Load Effect
                    if (File.Exists(xEffectInstances[iIndex].EffectFilename))
                    {
                        LoadEffect(xEffectInstances[iIndex], iIndex);
                    }
                    else
                    {
                        LoadDefaultEffect(iIndex);
                    }
                    #endregion
                }

                m_xMeshContainers.Add(m_xMeshContainer);

                #endregion

                m_bInitialised = true;
            }
        }

        protected Mesh NormalizeMesh(Mesh xMesh)
        {
            int[] adjacancy;
            xMesh = xMesh.Clone(
                MeshFlags.Managed,
                xMesh.VertexFormat | VertexFormats.Normal,
                m_xDevice
            );
            adjacancy = new int[xMesh.NumberFaces * 3];
            xMesh.GenerateAdjacency(0.01f, adjacancy);
            xMesh.ComputeNormals(adjacancy);
            return xMesh;
        }

        protected void LoadDefaultEffect(int iIndex)
        {
            try
            {
                String strFileName = "SoulReaver1Default.fx";

                Assembly xExecutingAssembly = Assembly.GetExecutingAssembly();
                Stream xStream              = xExecutingAssembly.GetManifestResourceStream(GetType(), strFileName);
                String strErrors;
                Effect xEffect              = Effect.FromStream(m_xDevice, xStream, null, ShaderFlags.None, null, out strErrors);
                Texture xTexture            = m_xMeshContainer.Textures[iIndex];
                Material xMaterial          = m_xMeshContainer.Materials[iIndex];

                #region Texture
                try
                {
                    EffectHandle xTexture0Handle = xEffect.GetParameter(null, "xTexture0");
                    xEffect.SetValue(xTexture0Handle, xTexture);
                }
                catch (Exception) { }
                #endregion

                #region Diffuse
                EffectHandle xDiffuseHandle = xEffect.GetParameter(null, "xDiffuse");
                float[] fDiffuse = { 1.0f, 1.0f, 1.0f, 1.0f };
                try
                {
                    fDiffuse = new float[]
                    {
                        xMaterial.Diffuse.R / 255,
                        xMaterial.Diffuse.G / 255,
                        xMaterial.Diffuse.B / 255,
                        xMaterial.Diffuse.A / 255
                    };
                }
                catch (Exception) {}
                xEffect.SetValue(xDiffuseHandle, fDiffuse);
                #endregion

                #region Lighting
                EffectHandle xLighting = xEffect.GetParameter(null, "xLighting");
                xEffect.SetValue(xLighting, DefaultLighting);
                #endregion

                #region ColorVertex
                EffectHandle xColorVertex = xEffect.GetParameter(null, "xColorVertex");
                xEffect.SetValue(xColorVertex, DefaultColorVertex);
                #endregion

                #region ColorOperation
                EffectHandle xColorOperation = xEffect.GetParameter(null, "xColorOperation");
                xEffect.SetValue(xColorOperation, (int)DefaultColorOperation);
                #endregion

                #region AmbientMaterialSource
                EffectHandle xAmbientMaterialSource = xEffect.GetParameter(null, "xAmbientMaterialSource");
                xEffect.SetValue(xAmbientMaterialSource, (int)DefaultAmbientMaterialSource);
                #endregion

                #region DiffuseMaterialSource
                EffectHandle xDiffuseMaterialSource = xEffect.GetParameter(null, "xDiffuseMaterialSource");
                xEffect.SetValue(xDiffuseMaterialSource, (int)DefaultDiffuseMaterialSource);
                #endregion

                m_xMeshContainer.Effects[iIndex] = xEffect;
                m_xMeshContainer.DefaultEffects[iIndex] = true;
            }
            catch (Exception)
            {
                return;
            }
        }

        protected void LoadEffect(EffectInstance xEffectInstance, int iIndex)
        {
            try
            {
                Effect xEffect = Effect.FromFile(m_xDevice, xEffectInstance.EffectFilename, null, ShaderFlags.None, null);

                foreach (EffectDefault xDefault in xEffectInstance.GetDefaults())
                {
                    EffectHandle xParamHandle = xEffect.GetParameter(null, xDefault.ParameterName);

                    switch (xDefault.Type)
                    {
                        case EffectDefaultType.Floats:
                            try
                            {
                                xEffect.SetValue(xParamHandle, (float[])xDefault.Data);
                            }
                            catch (Exception) { }
                            break;
                        case EffectDefaultType.Integer:
                            try
                            {
                                xEffect.SetValue(xParamHandle, (int)xDefault.Data);
                            }
                            catch (Exception) { }
                            break;
                        case EffectDefaultType.String:
                            try
                            {
                                xEffect.SetValue(xParamHandle, (String)xDefault.Data);
                            }
                            catch (Exception) { };
                            break;
                    }
                }

                m_xMeshContainer.Effects[iIndex] = xEffect;
                m_xMeshContainer.DefaultEffects[iIndex] = false;
            }
            catch (Exception) { }
        }

        public virtual void LoadTexture(String strFileName, int iTextureIndex)
        {
            LoadTexture(strFileName, m_iModelIndex, iTextureIndex);
        }

        public virtual void LoadTexture(String strFileName, int iModelIndex, int iTextureIndex)
        {
            try
            {
                ModelExMeshContainer xMeshContainer = m_xMeshContainers[iModelIndex];
                xMeshContainer.TextureNames[iTextureIndex] = Path.GetFileName(strFileName);
                xMeshContainer.Textures[iTextureIndex] = new Texture(m_xDevice, new Bitmap(strFileName), Usage.None, Pool.Managed);

                try
                {
                    if (xMeshContainer.DefaultEffects[iTextureIndex])
                    {
                        EffectHandle xTexture0Handle = xMeshContainer.Effects[iTextureIndex].GetParameter(null, "xTexture0");
                        xMeshContainer.Effects[iTextureIndex].SetValue(xTexture0Handle, xMeshContainer.Textures[iTextureIndex]);
                    }
                }
                catch (Exception) { }

                #region Old way of loading from a file...

                //m_axTextures[iIndex] = TextureLoader.FromFile(m_xDevice, strFileName);

                // Back up the texture to a stream... 
                //Bitmap xTextureBmp = new Bitmap(strFileName);
                //MemoryStream xTextureStream = new MemoryStream();
                //// xTextureStream.WriteTo(xOtherStream);
                //xTextureBmp.Save(xTextureStream, System.Drawing.Imaging.ImageFormat.Png);
                //m_axTextures[iIndex] = new Texture(m_xDevice, xTextureStream, Usage.None, Pool.Managed);

                #endregion

                return;
            }
            catch (Exception) { }

            try
            {
                ModelExMeshContainer xMeshContainer = m_xMeshContainers[iModelIndex];
                xMeshContainer.TextureNames[iTextureIndex] = "";
                xMeshContainer.Textures[iTextureIndex] = null;
                return;
            }
            catch (Exception) { }
        }

        public virtual void LoadTexturesFromFile(String strFileName)
        {
            try
            {
                for (int m = 0; m < m_xMeshContainers.Count; m++)
                {
                    LoadTexturesFromFile(strFileName, m);
                }
            }
            catch (Exception) { }
        }

        public virtual void LoadTexturesFromFile(String strFileName, int iModelIndex)
        {
        }

        public virtual void LoadTexturesFromFolder(String strFolderName)
        {
            try
            {
                for (int m = 0; m < m_xMeshContainers.Count; m++)
                {
                    LoadTexturesFromFolder(strFolderName, m);
                }
            }
            catch (Exception) { }
        }

        public virtual void LoadTexturesFromFolder(String strFolderName, int iModelIndex)
        {
            try
            {
                ModelExMeshContainer xMeshContainer = m_xMeshContainers[iModelIndex];
                for (int t = 0; t < xMeshContainer.TextureNames.Length; t++)
                {
                    String strFileName = "";
                    try
                    {
                        strFileName = Path.Combine(strFolderName, xMeshContainer.TextureNames[t]);
                    }
                    catch (Exception) { }
                    LoadTexture(strFileName, iModelIndex, t);
                }
            }
            catch (Exception) { }
        }

        public virtual void Save(String strFileName)
        {
            //for (int m = 0; m < m_xMeshContainers.Count; m++)
            //{
            //    int[] aiAdjacancy = new int[m_xMeshContainers[m].Mesh.NumberFaces * 3];
            //    m_xMeshContainers[m].Mesh.GenerateAdjacency(0.1f, aiAdjacancy);

            //    EffectInstance[] xEffectInstances = new EffectInstance[m_xMeshContainers[m].Mesh.NumberAttributes];
            //    for (int e = 0; e < xEffectInstances.Length; e++)
            //    {
            //        int iNumParameters = m_xMeshContainers[m].Effects[e].Description.Parameters;
            //        EffectDefault[] xEffectDefaults = new EffectDefault[iNumParameters];
            //        for (int p = 0; p < iNumParameters; p++)
            //        {
            //            EffectHandle xHandle = m_xMeshContainers[m].Effects[e].GetParameter(null, p);
            //            ParameterDescription xDescription = m_xMeshContainers[m].Effects[e].GetParameterDescription(xHandle);
            //            GraphicsStream xStream = m_xMeshContainers[m].Effects[e].GetValue(xHandle, xDescription.Bytes);
            //            byte[] acBuffer = new byte[xDescription.Bytes];
            //            xStream.Read(acBuffer, 0, xDescription.Bytes);

            //            xEffectDefaults[p] = new EffectDefault();
            //            xEffectDefaults[p].ParameterName = xDescription.Name;
            //            xEffectDefaults[p].Type = m_xMeshContainers[m].s
            //            xEffectDefaults[p].Data = acBuffer;
            //        }
            //        xEffectInstances[e].EffectFilename = "SoulReaverDefault.fx";
            //        xEffectInstances[e].SetDefaults(xEffectDefaults);
            //    }

            //    m_xMeshContainers[m].Mesh.Save(
            //        "Testing.x",
            //        aiAdjacancy,
            //        m_xMeshContainers[m].GetMaterials(),
            //        xEffectInstances,
            //        XFileFormat.Text
            //    );
            //}
        }

        public void ResetRotation()
        {
            m_xArcBall.Rotation = Quaternion.Identity;
        }

        public virtual void Render()
        {
            #region Update Orientation
            m_xRotation = m_xArcBall.Rotation;
            #endregion

            StateBackup xStateBackup = new StateBackup(ref m_xDevice);

            for (int i = 0; i < m_xMeshContainer.Effects.Length; i++)
            {
                int iPasses = m_xMeshContainer.Effects[i].Begin(FX.None);

                for (int p = 0; p < iPasses; p++)
                {
                    m_xMeshContainer.Effects[i].BeginPass(p);

                    #region Textures
                    if (m_bTextures)
                    {
                        m_xDevice.SetTexture(0, m_xMeshContainer.Textures[i]);
                    }
                    else
                    {
                        m_xDevice.SetTexture(0, null);
                    }
                    #endregion

                    #region Materials
                    m_xDevice.Material = m_xMeshContainer.Materials[i];
                    #endregion

                    #region Lighting
                    if (!m_bLighting)
                    {
                        m_xDevice.RenderState.Lighting = false;
                    }
                    #endregion

                    #region Texture Filtering
                    if (!m_bTextureFiltering)
                    {
                        m_xDevice.SamplerState[0].MagFilter = TextureFilter.Point;
                        m_xDevice.SamplerState[0].MinFilter = TextureFilter.Point;
                    }
                    #endregion

                    #region Cull Mode
                    m_xDevice.RenderState.CullMode = (Cull)m_eCullMode;
                    #endregion

                    #region Wireframe Mode
                    if (m_bWireframe)
                    {
                        m_xDevice.RenderState.FillMode = FillMode.WireFrame;
                    }
                    #endregion

                    #region Transforms
                    m_xDevice.Transform.World =
                        Matrix.Translation(Position - Centre) *
                        Matrix.Scaling(Scale, Scale, Scale) *
                        Matrix.RotationQuaternion(m_xRotation);
                    #endregion

                    #region Rendering
                    m_xMeshContainer.Mesh.DrawSubset(i);
                    #endregion

                    m_xMeshContainer.Effects[i].EndPass();
                }

                m_xMeshContainer.Effects[i].End();
            }

            xStateBackup.Restore();
        }
    }
}
